/// \file HostFileBdos.h
///
///  Virtual access to host filesystem via CP/Net-style I/O device.
///
///  \date Feb 19, 2016
///  \author Douglas Miller
///

#ifndef HOSTFILEBDOS_H_
#define HOSTFILEBDOS_H_

#include "NetworkServer.h"
#include "propertyutil.h"

#include <dirent.h>
#include <string>

#define BDOS_FUNC(name)							\
	int name(uint8_t * msgbuf, int len);				\
	static int name(HostFileBdos * thus, uint8_t * msgbuf, int len) {	\
		return thus->name(msgbuf, len);				\
	}

// An instance of HostFileBdos represents a client-server pair.
// This is typically a dedicated network connection (e.g. socket),
// but may also be an informal pairing such as a virtual machine's
// "connection" to the native host.

class HostFileBdos : public NetworkServer {
public:
	HostFileBdos(PropertyUtil::PropertyMapT &props,
			std::vector<std::string> args, uint8_t cltId);
	virtual ~HostFileBdos();

	virtual int checkRecvMsg(uint8_t clientId, uint8_t *msgbuf, int len);
	virtual int sendMsg(uint8_t *msgbuf, int len);

	static void initCfg(char tmp, uint8_t sid, uint8_t max);
private:
	char *dir;
	uint8_t serverId;
	uint8_t clientId;

	struct fcb {
		uint8_t drv;
		uint8_t name[11];
		uint8_t ext;
		uint8_t s1[2];
		uint8_t rc;
		union {
			uint8_t d0[16];
			struct {
				uint16_t fd;
				uint16_t fd_;
				int oflags;
			};
		};
		uint8_t cr;
		uint8_t rr[3];
	} __attribute__((packed));

	#define F1(fcb)	fcb->name[0]
	#define F2(fcb)	fcb->name[1]
	#define F3(fcb)	fcb->name[2]
	#define F4(fcb)	fcb->name[3]
	#define F5(fcb)	fcb->name[4]
	#define F6(fcb)	fcb->name[5]
	#define F7(fcb)	fcb->name[6]
	#define F8(fcb)	fcb->name[7]
	#define T1(fcb)	fcb->name[8]
	#define T2(fcb)	fcb->name[9]
	#define T3(fcb)	fcb->name[10]

	#define CLR_ATTR(attr)	(attr &= 0x7f)
	#define SET_ATTR(attr)	(attr |= 0x80)
	#define TEST_ATTR(attr)	((attr & 0x80) != 0)

	// These could be implemented using Unix file modes.
	#define ATTR_RO(fcb)	TEST_ATTR(T1(fcb))
	#define ATTR_SYS(fcb)	TEST_ATTR(T2(fcb))
	#define ATTR_AR(fcb)	TEST_ATTR(T3(fcb))
	#define SET_RO(fcb)	SET_ATTR(T1(fcb))
	#define SET_SYS(fcb)	SET_ATTR(T2(fcb))
	#define SET_AR(fcb)	SET_ATTR(T3(fcb))

	// Interface attributes, never stored to disk.
	// CP/M 3 interface attributes:
	#define ATTR_PC(fcb)	TEST_ATTR(F5(fcb))	// CLOSE: partial close
	#define ATTR_DELX(fcb)	TEST_ATTR(F5(fcb))	// DELETE: only delete XFCBs
	#define ATTR_PW(fcb)	TEST_ATTR(F6(fcb))	// MAKE: assign password
	#define ATTR_BC(fcb)	TEST_ATTR(F6(fcb))	// SET ATTR: set byte count
	// MP/M 3 interface attributes (in addition to (most) CP/M 3):
	#define ATTR_UNLK(fcb)	TEST_ATTR(F5(fcb))	// OPEN/MAKE: unlocked mode
	#define ATTR_OPRO(fcb)	TEST_ATTR(F6(fcb))	// OPEN: R/O mode

	#define SET_USR0(fcb)	SET_ATTR(F8(fcb))	// File from user 0 (cur not)

	// Compatability Attributes bit definitions
	static const uint8_t COMPAT_PRO = 0b10000000;	// Pseudo-R/O on OPEN
	static const uint8_t COMPAT_PC = 0b01000000;	// Partial CLOSE for all
	static const uint8_t COMPAT_NCC = 0b00100000;	// No checksum on CLOSE
	static const uint8_t COMPAT_NCS = 0b00010000;	// No checksum ever

	struct fcbdate {
		uint16_t date;
		uint8_t hour, min;
	} __attribute__((packed));

	struct cpmdate {
		uint16_t date;
		uint8_t hour, min, sec;
	} __attribute__((packed));

	struct sfcbs {
		struct fcbdate atim;
		struct fcbdate utim;
		uint8_t pwmode;
		uint8_t reserved;
	} __attribute__((packed));

	struct sfcb {
		uint8_t drv; // always 0x21
		struct sfcbs fcbs[3];
		uint8_t reserved;
	} __attribute__((packed));

	struct dirl {
		uint8_t drv;	// always 0x20
		uint8_t name[11];
		uint8_t ext;	// mode byte
		uint8_t s1[2];	// 0x00 0x00
		uint8_t rc;	// 0x00
		uint8_t pwd[8];	// password
		struct fcbdate ctim;
		struct fcbdate utim;
	} __attribute__((packed));

	struct search {
		int iter;
		int lastit;
		int endit;
		uint8_t maxdc;	// 3 or 4, depending on timestamps (SFCBs)
		uint8_t drv;
		uint8_t usr;
		uint8_t ext;
		uint8_t full;
		uint8_t cpm3;
		uint8_t bc;
		off_t size;
		mode_t mode;
		struct find {
			DIR *dir;
			char pat[16];
			char path[1024];
			int dirlen;
		} find;
	};

	struct dpb {
		uint16_t spt;
		uint8_t bsh;
		uint8_t blm;
		uint8_t exm;
		uint16_t dsm;
		uint16_t drm;
		uint8_t al0;
		uint8_t al1;
		uint16_t cks;
		uint16_t off;
	} __attribute__((packed));

	struct svrcfg {
		uint8_t tmp;
		uint8_t sts;
		uint8_t id;
		uint8_t max;
		uint8_t cur;
		uint16_t vec;
		uint8_t rid[16];
	} __attribute__((packed));

	struct OpenFile {
		int fd;
		uint16_t drv;
		uint16_t _res;
	};

	int lsts[16];

	static const int DEF_BLS_SH = 14;	// 2^14 = 16K
	static const int DEF_BLS = (1 << DEF_BLS_SH);
	static const int DEF_NBLOCKS = 128;	// keep alloc vec small, disk size 2M
	static const int DEF_NFILE = 32; // Probably never need even 8.
	static const uint8_t dirMode = 0b01100001;

	static int(*bdosFunctions[256])(HostFileBdos *, uint8_t *, int);

	static struct svrcfg cfgTab;
	static int addClient(uint8_t cid);
	static int rmClient(uint8_t cid);

	uint16_t curDma;
	void *curDmaReal;
	uint16_t curDsk;
	uint16_t curUsr;
	uint16_t curROVec;
	uint16_t curLogVec;
	uint8_t curCompat;
	mode_t uMask;
	struct search curSearch;
	struct dpb curDpb;
	int phyExt; // phy ext size, recs
	struct OpenFile openFiles[DEF_NFILE];
	char fileName[sizeof((struct dirent *)0)->d_name];
	char pathName[PATH_MAX];
	struct sfcb sFcb;

	BDOS_FUNC(listOut);
	BDOS_FUNC(getDPB);
	BDOS_FUNC(writeProt);
	BDOS_FUNC(resetDrive);
	BDOS_FUNC(getROVec);
	BDOS_FUNC(getAllocVec);
	BDOS_FUNC(getLoginVec);
	BDOS_FUNC(selectDisk);
	BDOS_FUNC(openFile);
	BDOS_FUNC(closeFile);
	BDOS_FUNC(searchNext);
	BDOS_FUNC(searchFirst);
	BDOS_FUNC(deleteFile);
	BDOS_FUNC(readSeq);
	BDOS_FUNC(writeSeq);
	BDOS_FUNC(createFile);
	BDOS_FUNC(renameFile);
	BDOS_FUNC(readRand);
	BDOS_FUNC(writeRand);
	BDOS_FUNC(setRandRec);
	BDOS_FUNC(compFileSize);
	BDOS_FUNC(setFileAttrs);
	BDOS_FUNC(accessDrive);
	BDOS_FUNC(freeDrive);
	BDOS_FUNC(writeRandZF);
	BDOS_FUNC(lockRec);
	BDOS_FUNC(unlockRec);
	BDOS_FUNC(getFreeSp);
	BDOS_FUNC(flushBuf);
	BDOS_FUNC(freeBlks);
	BDOS_FUNC(truncFile);
	BDOS_FUNC(login);
	BDOS_FUNC(logoff);
	BDOS_FUNC(setCompAttrs);
	BDOS_FUNC(getServCfg);
	BDOS_FUNC(setDefPwd);
	BDOS_FUNC(getDirLab);
	BDOS_FUNC(readFStamps);
	BDOS_FUNC(getTime);

	// Utility functions
	int cpmDrive(char *buf, int drive);
	int cpmFilename(char *buf, int user, char *file);
	int cpmPath(char *buf, int drive, int user, char *file);
	char *cpmNameFound(struct search::find *find);
	char *cpmPathFound(struct search::find *find);
	void cpmFindInit(struct search::find *find, int drive, char *pattern);
	const char *cpmFind(struct search::find *find, uint8_t usr);
	void getFileName(char *dst, struct fcb *fcb);
	void getAmbFileName(char *dst, struct fcb *fcb, uint8_t usr);
	void copyOutDir(uint8_t *dma, const char *name);
	int fullSearch(uint8_t *dirbuf, struct search *search, const char *ff);
	int copyOutSearch(uint8_t *buf, const char *name);
	const char *commonSearch(struct search *search);
	const char *doSearch(struct search *search);
	const char *startSearch(struct fcb *fcb, struct search *search, uint8_t u);
	void endSearch(struct search *search);
	void seekFile(struct fcb *fcb);
	int openFileFcb(struct fcb *fcb, uint8_t u);
	int newFileFcb();
	int locFileFcb(struct fcb *fcb);
	void putFileFcb(struct fcb *fcb, int ix, int fd, uint8_t drv);
	int getFileFcb(struct fcb *fcb);
	int getFileWrFcb(struct fcb *fcb);
	int closeFileFcb(struct fcb *fcb);
	int makeFileFcb(struct fcb *fcb, uint8_t u);
	void unix2cpmdate(time_t unx, struct cpmdate *cpm);
	int recLocking(uint8_t *msgbuf, int len, int mode);
	void closeAll(uint16_t vec);
	void clearIntfAttrs(struct fcb *fcb);
	int getFileStat(struct fcb *fcb, uint8_t u, struct stat *stb);

	inline uint8_t getDrv(struct fcb *fcb) {
		uint8_t d = fcb->drv & 0x1f;
		if (!d) {
			d = curDsk;
		} else {
			--d;
		}
		return d & 0x0f;
	}
};

#endif	// HOSTFILEBDOS_H_
