/// \file cpnet_server.cpp
///
/// \date Apr 23, 2016
/// \author Douglas Miller
///

#include <iostream>
#include <signal.h>
#include <stdlib.h>
#include <string>
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>

#include "FT245RServer.h"
#include "propertyutil.h"
#include "host_logging.h"

void help(char *argv0, int xit) {
	fprintf(stderr, "Usage: %s -i node-id [-s node-id] [options]\n", argv0);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "        -i node-id   Node ID to assign to client/requester\n");
	fprintf(stderr, "        -s node-id   Node ID for this server instance (default 00)\n");
	fprintf(stderr, "        -d path      Set root dir to <path>\n");
	fprintf(stderr, "        -l path      Log to <path> file instead of stderr\n");
	fprintf(stderr, "        -P #:string  Use <string> for printer <#> (0-F)\n");
	fprintf(stderr, "        -t tty       Use <tty> for client communication\n");
	fprintf(stderr, "        -T letter    Set temp drive to <letter> (A-P)\n");
	fprintf(stderr, "This programs uses server ID 00\n");
	fprintf(stderr, "Env vars \"hostfilebdos*\" will be used as properties\n");
	exit(xit);
}

FILE *__log__ = stderr;

int main(int argc, char **argv, char **env) {
	int c;
	unsigned long n;
	char *e;
	extern char *optarg;
	PropertyUtil::PropertyMapT props;
	struct termios termios0;
	struct termios termios;
	int ret = -1;
	int rxfd = -1;
	int txfd = -1;
	char prop[32];

	for (char **ep = env; *ep != NULL; ++ep) {
		// property/env-var name format:
		//     hostfilebdos<sid>_<prop>
		// where <sid> is 2-digit hex server ID
		if (strncmp(*ep, "hostfilebdos", 12) == 0) {
			e = strchr(*ep, '=');
			if (e == NULL) {
				continue;
			}
			*e++ = '\0';
			props[*ep] = e;
		}
	}

	std::string hfb = "HostFileBdos";
	char *dir = NULL;
	char tmp = '\0';
	char *tty = NULL;
	uint8_t cid = 0xfe;
	uint8_t sid = 0x00;
	if (argc < 2) {
		help(argv[0], 0);
	}
	while ((c = getopt(argc, argv, "d:i:l:P:s:t:T:")) != EOF) {
		switch (c) {
		case 'd':
			dir = optarg;
			break;
		case 'i':
			n = strtoul(optarg, &e, 16);
			if (*e != '\0' || n < 1 || n >= 255) {
				fprintf(stderr, "Invalid client ID: %s\n", optarg);
				return 1;
			}
			cid = n & 0x0ff;
			break;
		case 'l':
			__log__ = fopen(optarg, "w");
			if (__log__ == NULL) {
				perror(optarg);
				return 1;
			}
			break;
		case 'P':
			if (!isxdigit(optarg[0]) || optarg[1] != ':') {
				fprintf(stderr, "Invalid printer: %s\n", optarg);
				return 1;
			}
			sprintf(prop, "hostfilebdos%02x_lst%c", sid, toupper(optarg[0]));
			props[prop] = &optarg[2];
			break;
		case 's':	// must appear before rest (at least -P)
			n = strtoul(optarg, &e, 16);
			if (*e != '\0' || n < 0 || n >= 255) {
				fprintf(stderr, "Invalid server ID: %s\n", optarg);
				return 1;
			}
			sid = n & 0x0ff;
			break;
		case 't':
			tty = optarg;
			break;
		case 'T':
			if (!isalpha(optarg[0]) || (optarg[1] != '\0' &&
					(optarg[1] != ':' || optarg[2] != '\0'))) {
				fprintf(stderr, "Invalid temp drive: %s\n", optarg);
				return 1;
			}
			tmp = toupper(optarg[0]);
			break;
		default:
			fprintf(stderr, "Unknown option -%c\n", c);
			help(argv[0], 1);
			break;
		}
	}
	if (tty == NULL) {
		setbuf(stdin, NULL);
		rxfd = 0;	// or dup()?
		txfd = 1;
	} else {
		rxfd = txfd = open(tty, O_RDWR);
		if (rxfd < 0) {
			perror(tty);
			return 1;
		}
	}
	if (isatty(rxfd)) {
		ret = tcgetattr(rxfd, &termios);
		if (ret == 0) {
			memcpy(&termios0, &termios, sizeof(termios0));
			termios.c_lflag &= ~ICANON;
			termios.c_lflag &= ~ECHO;
			ret = tcsetattr(rxfd, 0, &termios);
		}
	}
	if (dir == NULL) {
		hfb += " ";
		hfb += getenv("HOME");
		hfb += "/HostFileBdos";
	} else {
		hfb += " ";
		hfb += dir;
	}
	if (tmp != '\0') {
		hfb += " ";
		hfb += tmp;
	}
	sprintf(prop, "cpnet_server%02x", sid);
	props[prop] = hfb;

	FT245RServer *srv = new FT245RServer(props, "cpnet_", cid);
	srv->mainThread(rxfd, txfd);
	if (__log__ != stderr) {
		fclose(__log__);
	}

	if (ret == 0) {
		tcsetattr(rxfd, 0, &termios0);
	}

	return 0;
}
